from RDG import RDG, Node, Edge

def test_inserting_orfs_edges():
    g = RDG()
    g = RDG.load_example(g)
    node_key = g.get_new_node_key()

    start_codon_position = 15
    stop_codon_position = 25
    start_node = Node(key=node_key, node_type="start_codon", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

    node_key = node_key + 1
    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
    
    g.insert_ORF(g.edges[2], start_node, stop_node)
    assert list(g.edges.keys()) == [1,2,3,4,5,6,7]


def test_insterting_orfs_nodes():
    g = RDG()
    g = RDG.load_example(g)
    node_key = g.get_new_node_key()

    start_codon_position = 15
    stop_codon_position = 25
    start_node = Node(key=node_key, node_type="start_codon", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

    node_key = node_key + 1
    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
    
    g.insert_ORF(g.edges[2], start_node, stop_node)
    assert list(g.nodes.keys()) == [1,2,3,4,5,6,7,8]


def test_insterting_orfs_branch_points():
    g = RDG()
    g = RDG.load_example(g)
    node_key = g.get_new_node_key()

    start_codon_position = 15
    stop_codon_position = 25
    start_node = Node(key=node_key, node_type="start_codon", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

    node_key = node_key + 1
    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
    
    g.insert_ORF(g.edges[2], start_node, stop_node)

    branch_points = g.get_branch_points()

    assert len(branch_points) == 2


def test_add_open_reading_frame():
    g = RDG()
    g = RDG.load_example(g)

    g.add_open_reading_frame(15,25)
    branch_points = g.get_branch_points()

    assert len(branch_points) == 2


from RDG import RDG, Node, Edge
from sqlitedict import SqliteDict


def to_dict(obj):
    output ={}
    print(obj, type(obj), obj.__dict__.items())
    for key, item in obj.__dict__.items():
        if isinstance(item, list):
            l = []
            for item in item:
                if isinstance(item, int):
                    l.append(item)
                else:
                    d = to_dict(item)
                    l.append(d)
            output[key] = l
        else:
            output[key] = item

    return output


def save(graph, save_file):

    transcript = graph.locus
    graph_dict = {transcript:{}}
    for attr, value in graph.__dict__.items():
        if attr not in graph_dict[transcript]:
            graph_dict[transcript][attr] = {}

        if isinstance(value, dict):
            for key in value:
                graph_dict[transcript][attr][key] = to_dict(value[key])
        else:
            graph_dict[transcript][attr] = value

    try:  
        with SqliteDict(save_file) as output_dict:
            for key in graph_dict:
                output_dict[key] = graph_dict[key]
                output_dict.commit() 

    except Exception as ex:
        print("Error during storing data (Possibly unsupported):", ex)


def load(locus, cache_file="test_output.sqlite"):
    try:
        with SqliteDict(cache_file) as mydict:
            graph_dict = mydict[locus] # No need to use commit(), since we are only loading data!
    except Exception as ex:
        print("Error during loading data:", ex)
    
    dg = RDG()
    nodes = {}
    for key in graph_dict['nodes']:
        nodes[key] = Node(key, 
        graph_dict['nodes'][key]['node_type'], 
        graph_dict['nodes'][key]['node_start'],
        graph_dict['nodes'][key]['input_edges'],
        graph_dict['nodes'][key]['output_edges'],
        graph_dict['nodes'][key]['input_nodes'],
        graph_dict['nodes'][key]['output_nodes'])

    edges = {}

    for key in graph_dict['edges']:
        edges[key] = Edge(key, 
        graph_dict['edges'][key]['edge_type'], 
        graph_dict['edges'][key]['from_node'],
        graph_dict['edges'][key]['to_node'],
        graph_dict['edges'][key]['coordinates'])

    dg = dg.Load(locus_name=locus, locus_start=graph_dict['locus_start'], locus_stop=graph_dict['locus_stop'], nodes=nodes, edges=edges)
    return dg




# if __name__ == "__main__":
#     dg = RDG()
#     dg.add_open_reading_frame(30, 90)
#     dg.add_open_reading_frame(131, 171)
#     dg.add_open_reading_frame(150,850)
   
#     # save(dg, 'test_output.sqlite')
#     dg2 = load('', 'test_output.sqlite')

#     a = dg.describe()
#     print()
#     b = dg2.describe()
#     print(a == b)


class RDG(object):
    def __init__(self, locus_stop=1000, nodes=None, edges=None, name=None):
        """
        initialise a fresh empty RDG 
        """
        self.locus = name
        self.locus_start = 0
        self.locus_stop = locus_stop

        if nodes: 
            raise ValueError("A value for 'nodes' was provided. If you want to specify from an existing structure use RDG.load()")
        if edges: 
            raise ValueError("A value for 'edges' was provided. If you want to specify from an existing structure use RDG.load()")

        transcript_length = locus_stop - self.locus_start

        nodes = {
                1:Node(key=1, node_type="5_prime", coordinates=0, edges_out=[1], nodes_out=[2]),
                2:Node(key=2, node_type="3_prime", coordinates=transcript_length -1, edges_in=[1], nodes_in=[1])
            }

        edges = {1: Edge(1, "untranslated", from_node=1, to_node=2, coordinates=(1, transcript_length -1))}


        self.edges = edges
        self.nodes = nodes


    def Load(self, locus_name="unnamed", locus_start=0, locus_stop=1000, nodes=None, edges=None):
        """
        create a RDG from the given paramaters 
        """
        if nodes:
            self.nodes = nodes

        if edges:
            self.edges = edges 
        
        self.locus = locus_name
        self.locus_start = locus_start
        self.locus_stop = locus_stop
        return self

    def load_example(self):
        '''
        load a basic graph with one ORF 
        locus length = 1000
        orf coordinates = 10, 100
        '''
        nodes = {
            1:Node(key=1, node_type="5_prime", coordinates=0, edges_out=[1], nodes_out=[3]),
            2:Node(key=2, node_type="3_prime", coordinates=1000 -1, edges_in=[2], nodes_in=[3]),
            3:Node(key=3, node_type="start", coordinates=10, edges_in=[1], edges_out=[2,3], nodes_in=[1], nodes_out=[2,4]),
            4:Node(key=4, node_type="stop", coordinates=100, edges_in=[3], edges_out=[4], nodes_in=[3], nodes_out=[5]),
            5:Node(key=5, node_type="3_prime", coordinates=1000 -1, edges_in=[4], nodes_in=[4])
            }
        edges = {
            1:Edge(1, "untranslated", from_node=1, to_node=3, coordinates=(1, 10 -1)),
            2:Edge(2, "untranslated", from_node=3, to_node=2, coordinates=(11, 1000 -1)),
            3:Edge(3, "translated", from_node=3, to_node=4, coordinates=(11, 100)),
            4:Edge(4, "untranslated", from_node=4, to_node=5, coordinates=(101, 1000 -1))
        }
        g = RDG()
        g = RDG.Load(g, nodes=nodes, edges=edges)
        return g


    def vertices(self):
        """
        return a list of the vertices/nodes from the graph
        """
        return list(self.nodes.keys())


    def get_edges(self):
        '''
        return a list of keys of edges in the graph
        '''
        return list(self.edges.keys())


    def get_edges_from_to(self):
        '''
        return a list of edges (from, to) in the graph
        '''
        from_to = {} 
        for edge in self.edges.keys():
            from_to[(self.edges[edge].from_node, self.edges[edge].to_node)] = edge

        return from_to


    def get_new_node_key(self):
        """
        check the nodes to find the lowest available key
        """
        keys = self.nodes.keys()
        if len(keys) > 0:
            return max(keys) + 1
        else:
            return 1


    def get_new_edge_key(self):
        """
        check the edges to find the lowest available key
        """
        keys = self.edges.keys()
        if len(keys) > 0:
            return max(keys) + 1
        else:
            return 1

    def get_key_from_position(self, position, node_type):
        '''
        return the node key for the node of specified type at the stated position 
        '''
        for node in self.nodes:
            if self.nodes[node].node_start == position and self.nodes[node].node_type == node_type:
                return node
        raise ValueError(f"There is no stop codon at the locus ({position}) specified for readthrough")



    def remove_edge(self, edge_key):
        '''
        remove the edge with the given key from the graph Also removing references from adjacent nodes 
        '''
        from_node = self.edges[edge_key].from_node
        to_node = self.edges[edge_key].to_node
        if edge_key in self.nodes[from_node].output_edges:
            self.nodes[from_node].output_edges.remove(edge_key)
        
        if to_node in self.nodes[from_node].output_nodes:
            self.nodes[from_node].output_nodes.remove(to_node)

        if edge_key in self.nodes[to_node].input_edges:
            self.nodes[to_node].input_edges.remove(edge_key)

        if from_node in self.nodes[to_node].input_nodes:
            self.nodes[to_node].input_nodes.remove(from_node)

        if edge_key in self.edges:
            self.edges.pop(edge_key)


    def add_edge(self, edge, from_node_key, to_node_key):
        '''
        reintroduce edges and references with the new edge 
        '''
        self.nodes[from_node_key].output_edges.append(edge.key)
        self.nodes[from_node_key].output_nodes.append(to_node_key)

        self.nodes[to_node_key].input_edges.append(edge.key)
        self.nodes[to_node_key].input_nodes.append(from_node_key)

        self.edges[edge.key] = edge 
        self.edges[edge.key].frame = self.nodes[from_node_key].node_start%3



    def add_node(self, node):
        self.nodes[node.key] = node


    def update_edge(self, edge_key, new_from_node, new_to_node, new_coordinates):
        '''
        functionality for changing 
        when inserting an orf into an untranslated edge we update the untranslated edge to now start at the new start codon instead

        '''
        old_from_node = self.edges[edge_key].from_node
        old_to_node = self.edges[edge_key].to_node

        if old_from_node != new_from_node:
            if edge_key in self.nodes[old_from_node].output_edges:
                self.nodes[old_from_node].output_edges.remove(edge_key)
            
            if edge_key not in self.nodes[new_from_node].output_edges:
                self.nodes[new_from_node].output_edges.append(edge_key)
            
            if old_to_node in self.nodes[old_from_node].output_nodes:
                self.nodes[old_from_node].output_nodes.remove(old_to_node)
            
            if new_from_node not in self.nodes[new_from_node].output_nodes:
                self.nodes[new_from_node].output_nodes.append(new_to_node)
        


        if old_to_node != new_to_node:
            if edge_key in self.nodes[old_to_node].input_edges:
                self.nodes[old_to_node].input_edges.remove(edge_key)

            if edge_key not in self.nodes[new_to_node].input_edges:
                self.nodes[new_to_node].input_edges.append(edge_key)

            if old_from_node in self.nodes[old_to_node].input_nodes:
                self.nodes[old_to_node].input_nodes.remove(old_from_node)

            if new_from_node in self.nodes[new_to_node].input_nodes:
                self.nodes[new_to_node].input_nodes.remove(new_from_node)
        self.edges[edge_key].coordinates = new_coordinates
        self.edges[edge_key].from_node = new_from_node


    def insert_ORF(self, edge, start_node, stop_node):
        '''
        handle inserting node into just one edge including down path 
        '''
        five_prime_edge_coords = (edge.coordinates[0], start_node.node_start - 1)
        coding_edge_coords = (start_node.node_start, stop_node.node_start)
        three_prime_edge_coords = (stop_node.node_start + 1, edge.coordinates[1])

        edge_key = self.get_new_edge_key()
        five_prime = Edge(key=edge_key, edge_type="untranslated", from_node=edge.from_node, to_node=start_node.key, coordinates=five_prime_edge_coords)

        self.add_node(start_node)
        self.add_edge(five_prime, edge.from_node, start_node.key)

        edge_key = self.get_new_edge_key()
        coding = Edge(key=edge_key, edge_type="translated", from_node=start_node.key, to_node=stop_node.key, coordinates=coding_edge_coords)

        self.add_node(stop_node)
        self.add_edge(coding, start_node.key, stop_node.key)


        terminal_node_key = self.get_new_node_key()
        terminal_node = Node(key=terminal_node_key, node_type="3_prime", coordinates=edge.coordinates[1], edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

        edge_key = self.get_new_edge_key()
        three_prime = Edge(key=edge_key, edge_type="untranslated", from_node=stop_node.key, to_node=terminal_node.key, coordinates=three_prime_edge_coords)

        self.add_node(terminal_node)
        self.add_edge(three_prime, stop_node.key, terminal_node.key)

        self.update_edge(edge.key, start_node.key, edge.to_node, (start_node.node_start, self.nodes[edge.to_node].node_start))


    def is_input_edge_translated(self, node):
        '''
        return boolean as to whether an edge entering this node is translated 
        '''
        input_edges = self.nodes[node].input_edges
        boolean = False
        for edge in input_edges:
            if self.edges[edge].edge_type == 'translated':
                boolean = True
        
        return boolean


    def root_to_node_of_acyclic_node_path(self, node):
        '''
        return path from root to given node as a list of node keys
        '''
        path = [node]
        reached_root = False

        while not reached_root: 
            in_node = self.nodes[node].input_nodes[0]
            path.append(in_node)
            if self.nodes[in_node].input_nodes == []:
                reached_root = True
            else:
                node = in_node
        return(path[::-1])


    def root_to_node_of_acyclic_edge_path(self, node):
        '''
        return path from root to given node as a list of edge keys
        '''
        if self.nodes[node].input_nodes == []:
            reached_root = True
            return []

        path = [self.nodes[node].input_edges[0]]
        reached_root = False

        while not reached_root: 
            in_node = self.nodes[node].input_nodes[0]
            if self.nodes[in_node].input_nodes == []:
                reached_root = True
            else:
                path.append(self.nodes[in_node].input_edges[0])
                node = in_node
        return path[::-1]
            

    def check_translation_upstream(self, from_node, upstream_limit=1):
        '''
        look upstream of an edges from node and see if any edges are of type 'translated' 
        used in reinit functionality. Upstream limit refers to the number of ORFs allowed upstream. ie. can the be multiple reinitiation events or just one etc
        '''
        edge_path_to_node_from_root = self.root_to_node_of_acyclic_edge_path(from_node)
        number_of_translated_regions = 0
        for edge in edge_path_to_node_from_root:
            if self.edges[edge].edge_type == 'translated':
                number_of_translated_regions += 1

        if number_of_translated_regions <= upstream_limit:
            return False
        return True


    def add_open_reading_frame(self, start_codon_position, stop_codon_position, reinitiation=False, upstream_limit=0):
        '''
        Handles all operations related to adding a new decision to the graph. Includes objects in graph and corrects all objects involved
        '''
        exisiting_edges = self.get_edges()
        clashing_edges = [] 
        for edge in exisiting_edges:
            if start_codon_position in range(self.edges[edge].coordinates[0], self.edges[edge].coordinates[1]) and self.edges[edge].edge_type != "translated":
                upstream_node = self.edges[edge].from_node
                clashing_edges.append((edge, upstream_node))
                
        
        for edge, upstream_node in clashing_edges:
            node_key = self.get_new_node_key()
            start_node = Node(key=node_key, node_type="start", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

            node_key = node_key + 1
            stop_node = Node(key=node_key, node_type="stop", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
            if reinitiation or not self.check_translation_upstream(upstream_node, upstream_limit=upstream_limit):
                self.insert_ORF(self.edges[edge], start_node, stop_node)

    
    def add_stop_codon_readthrough(self, readthrough_codon_position, next_stop_codon_position):
        '''
        Handles all operations related to adding a new decision to the graph at a stop codon. Includes objects in graph and corrects all objects involved

        '''
        readthrough_codon_key = self.get_key_from_position(readthrough_codon_position, "stop")
        three_prime_terminal_key = self.nodes[readthrough_codon_key].output_nodes[0]

        node_key = self.get_new_node_key()
        terminal_node_key = node_key + 1

        edge_key = self.get_new_edge_key()
        edge_key2 = edge_key + 1

        new_stop_node = Node(key=node_key, node_type="stop", coordinates=next_stop_codon_position, edges_in=[edge_key], edges_out=[edge_key2], nodes_in=[readthrough_codon_key], nodes_out=[terminal_node_key])
        self.add_node(new_stop_node)

        coding = Edge(key=edge_key, edge_type="translated", from_node=readthrough_codon_key, to_node=new_stop_node.key, coordinates=(self.nodes[readthrough_codon_key].node_start, next_stop_codon_position))
        self.add_edge(coding, readthrough_codon_key, new_stop_node.key)

        self.nodes[readthrough_codon_key].output_edges.append(coding.key)
        self.nodes[readthrough_codon_key].output_nodes.append(new_stop_node.key)
        self.nodes[readthrough_codon_key].node_type = "readthrough_stop"
        
        three_prime = Edge(key=edge_key2, edge_type="untranslated", from_node=new_stop_node.key, to_node=terminal_node_key, coordinates=(new_stop_node.node_start, self.nodes[three_prime_terminal_key].node_start))

        terminal_node_key = self.get_new_node_key()
        terminal_node = Node(key=terminal_node_key, node_type="3_prime", coordinates=self.nodes[three_prime_terminal_key].node_start, edges_in=[edge_key2], edges_out=[], nodes_in=[new_stop_node.key], nodes_out=[])
        self.add_node(terminal_node)

        self.add_edge(three_prime, new_stop_node.key, terminal_node_key)


    def add_frameshift(self, fs_position, next_stop_codon_position, shift):
        '''
        add a frameshifting event 
        '''
        exisiting_edges = self.get_edges()
        clashing_edges = [] 
        for edge in exisiting_edges:
            if fs_position in range(self.edges[edge].coordinates[0], self.edges[edge].coordinates[1]) and self.edges[edge].edge_type == "translated":
                upstream_node = self.edges[edge].from_node
                clashing_edges.append((edge, upstream_node))  

        for edge, upstream_node in clashing_edges:
            # edge and upstrream are keys
            shift_node_key = self.get_new_node_key()
            new_stop_node_key = shift_node_key + 1
            old_stop_node_key = self.edges[edge].to_node
            new_terminal_key = new_stop_node_key + 1

            old_stop_edge_key = self.get_new_edge_key()
            new_stop_edge_key = old_stop_edge_key + 1
            new_three_prime_edge_key = new_stop_edge_key + 1

            shift_node = Node(key=shift_node_key, 
            node_type="frameshift", 
            coordinates=fs_position, 
            edges_in=[edge], 
            edges_out=[old_stop_edge_key, new_stop_edge_key], 
            nodes_in=[upstream_node], 
            nodes_out=[old_stop_node_key, new_stop_node_key])
            self.add_node(shift_node)

            self.nodes[upstream_node].output_nodes.remove(old_stop_node_key)
            self.nodes[upstream_node].output_nodes.append(shift_node.key)

            self.edges[edge].to_node = shift_node_key

            self.nodes[old_stop_node_key].input_edges.remove(edge)
            self.nodes[old_stop_node_key].input_edges.append(old_stop_edge_key)
            self.nodes[old_stop_node_key].input_nodes.remove(upstream_node)
            self.nodes[old_stop_node_key].input_nodes.append(shift_node_key)

            
            new_stop_node = Node(key=new_stop_node_key, 
            node_type="stop", 
            coordinates=next_stop_codon_position, 
            edges_in=[new_stop_edge_key], 
            edges_out=[new_three_prime_edge_key], 
            nodes_in=[shift_node_key], 
            nodes_out=[new_terminal_key])
            self.add_node(new_stop_node)
            
            new_terminal_node = Node(key=new_terminal_key, 
            node_type="3_prime",
            coordinates=self.locus_stop, 
            edges_in=[new_three_prime_edge_key], 
            edges_out=[], 
            nodes_in=[new_stop_node_key], 
            nodes_out=[])
            self.add_node(new_terminal_node)

            old_stop_edge = Edge(key=old_stop_edge_key,
            edge_type="translated",
            from_node=shift_node_key,
            to_node=old_stop_node_key,
            coordinates=(shift_node.node_start, self.nodes[old_stop_node_key].node_start))
            self.add_edge(old_stop_edge, from_node_key=shift_node_key, to_node_key=old_stop_node_key)
            self.edges[old_stop_edge_key].frame = self.edges[edge].frame

            new_stop_edge = Edge(key=new_stop_edge_key,
            edge_type="translated",
            from_node=shift_node_key,
            to_node=new_stop_node_key,
            coordinates=(shift_node.node_start, self.nodes[new_stop_node_key].node_start))
            self.add_edge(new_stop_edge, from_node_key=shift_node_key, to_node_key=new_stop_node_key)


            new_three_prime_edge = Edge(key=new_three_prime_edge_key,
            edge_type="untranslated",
            from_node=new_stop_node_key,
            to_node=new_terminal_key,
            coordinates=(new_stop_node.node_start, new_terminal_node.node_start))
            self.add_edge(new_three_prime_edge, from_node_key=new_stop_node_key, to_node_key=new_terminal_key)



    def get_branch_points(self):
        '''
        return a list of nodes that are branch points
        '''
        branch_points = []
        for node in self.nodes:
            if len(self.nodes[node].output_edges) > 1:
                branch_points.append(node)
        return branch_points


    def get_endpoints(self):
        '''
        return list of terminal node keys 
        '''
        endpoints = [] 
        for node in self.nodes:
            if len(self.nodes[node].output_edges) == 0 and self.nodes[node].node_type == "3_prime":
                endpoints.append(node)
        return endpoints


    def get_startpoints(self):
        '''
        return list of start node keys 
        '''
        startpoints = [] 
        for node in self.nodes:
            if len(self.nodes[node].input_edges) == 0 and self.nodes[node].node_type == "5_prime":
                startpoints.append(node)
        return startpoints
    

    def get_start_nodes(self):
        '''
        return translation start keys
        '''
        translation_starts = [] 
        for node in self.nodes:
            if len(self.nodes[node].output_edges) == 2 and self.nodes[node].node_type == "start":
                translation_starts.append(node)
        return translation_starts


    def get_stop_nodes(self):
        '''
        return translation start keys
        '''
        translation_stops = [] 
        for node in self.nodes:
            if "stop" in  self.nodes[node].node_type:
                translation_stops.append(node)
        return translation_stops

    
    def get_orfs(self):
        '''
        return a list of coordinates (start, stop) describing the ORFs in the graph
        ''' 
        translation_starts = self.get_start_nodes()

        orfs = []
        for start in translation_starts:
            downstream_nodes = self.nodes[start].output_nodes

            for candidate_stop in downstream_nodes:
                if self.nodes[candidate_stop].node_type == "stop":
                    orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))

                elif self.nodes[candidate_stop].node_type == "frameshift":
                    # orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))
                    fs_downstream_nodes = self.nodes[candidate_stop].output_nodes
                    for new_candidate_stop in fs_downstream_nodes:
                        if self.nodes[new_candidate_stop].node_type == "stop":
                            orf = (self.nodes[start].node_start, self.nodes[candidate_stop].node_start, self.nodes[new_candidate_stop].node_start)
                            if orf not in orfs:
                                orfs.append(orf)

                elif self.nodes[candidate_stop].node_type == "readthrough_stop":
                    orfs.append((self.nodes[start].node_start, self.nodes[candidate_stop].node_start))
                    readthrough_downstream_nodes = self.nodes[candidate_stop].output_nodes

                    for new_candidate_stop in readthrough_downstream_nodes:
                        if self.nodes[new_candidate_stop].node_type == "stop":
                            orf = (self.nodes[start].node_start, self.nodes[new_candidate_stop].node_start)
                            if orf not in orfs:
                                orfs.append(orf)


        
        return orfs


    def get_frameshifts(self):
        '''
        return a list of keys for all frameshift nodes
        '''
        frameshifts = [] 
        for node in self.nodes:
            if "frameshift" in  self.nodes[node].node_type:
                frameshifts.append(node)
        return frameshifts


    def print_paths(self):
        '''
        read out the paths from the graph object 
        '''
        root_nodes = [] 
        for node in self.nodes:
            if self.nodes[node].node_type == "5_prime":
                root_nodes.append(node)

        all_paths = [] 
        for root in root_nodes:
            path_from_root_nodes = [root] 
            if self.nodes[root].output_nodes:
                for node in self.nodes[root].output_nodes:
                    self.explore_node_ouputs(node, all_paths, path_from_root_nodes)

            else:
                all_paths.append(path_from_root_nodes)
        print("all paths", all_paths)


    def statistics(self):
        stats = {}
        nodes = list(self.nodes.keys())
        calc_dict = {
            "edges": [],
            "frames": [],
            "types": [],
            "frames_freq": {},
            "types_freq": {},
        }
        
        calc_dict["edges"] = self.get_edges_from_to().keys()

        for i in nodes:
            calc_dict["frames"].append(self.nodes[i].frame)
            calc_dict["types"].append(self.nodes[i].node_type)

        for item in calc_dict["frames"]:
            if item in calc_dict["frames_freq"]:
                calc_dict["frames_freq"][item] += 1
            else:
                calc_dict["frames_freq"][item] = 1

        for item in calc_dict["types"]:
            if item in calc_dict["types_freq"]:
                calc_dict["types_freq"][item] += 1
            else:
                calc_dict["types_freq"][item] = 1

        for item in calc_dict["types_freq"]:
            key = "Number_of_" + item
            stats[key] = calc_dict["types_freq"][item]

        for item in calc_dict["frames_freq"]:
            key = "Number_of_nodes_in_frame_" + str(item)
            stats[key] = calc_dict["frames_freq"][item]


        stats["Node_keys"] = nodes
        stats["Number_of_nodes"] = len(self.nodes.keys())
        stats["Edges_keys"] = calc_dict["edges"]
        stats["Number_of_edges"] = len(calc_dict["edges"])
        stats["Number_of_node_types"] = len(calc_dict["types_freq"])
        return stats


    def describe(self):
        """
        print in the terminal a description of the graph
        """
        output = ""
        stats = self.statistics()
        for entry in stats:
            if "Number" in entry: 
                string = str(entry) + "\t" + str(stats[entry])
                output = output  + "\n" + string 
        return output







class Node(object):
    def __init__(self, key, node_type, coordinates, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[]):
        self.key = key
        self.node_type = node_type
        self.input_edges = edges_in
        self.output_edges = edges_out
        self.input_nodes = nodes_in 
        self.output_nodes = nodes_out 
        self.node_start = coordinates
        self.frame = self.node_start % 3


    def add_neighbour(self, neighbour, weight=0):
        self.connected[neighbour] = weight

    def get_connections(self):
        return self.connected.keys()

    def node_key(self):
        return self.key

    def node_type(self):
        return self.node_type




class Edge(object):
    def __init__(self, key, edge_type, from_node, to_node, coordinates):
        self.key = key
        self.edge_type = edge_type
        self.from_node = from_node
        self.to_node = to_node
        self.coordinates = coordinates

    def edge_key(self):
        return self.key

    def edge_type(self):
        return self.edge_type

    def from_node(self):
        return self.from_node

    def to_node(self):
        return self.to_node 
    
    def coordinates(self):
        return self.coordinates

    def frame(self):
        return self.coordinates[0] % 3 



# if __name__ == "__main__":

#     dg = RDG()
#     dg = dg.load_example()
#     dg.add_open_reading_frame(30, 90)

#     dg.add_open_reading_frame(150, 171)
#     dg.add_stop_codon_readthrough(90, 120)
#     print(dg.get_orfs())



from networkx.exception import AmbiguousSolution
from RDG import RDG, Node, Edge, plot, save
from pprint import pprint
from progress.bar import Bar
from time import time



# table = """TTT F      CTT L      ATT I      GTT V
# TTC F      CTC L      ATC I      GTC V
# TTA L      CTA L      ATA I      GTA V
# TTG L      CTG L      ATG M      GTG V
# TCT S      CCT P      ACT T      GCT A
# TCC S      CCC P      ACC T      GCC A
# TCA S      CCA P      ACA T      GCA A
# TCG S      CCG P      ACG T      GCG A
# TAT Y      CAT H      AAT N      GAT D
# TAC Y      CAC H      AAC N      GAC D
# TAA Stop   CAA Q      AAA K      GAA E
# TAG Stop   CAG Q      AAG K      GAG E
# TGT C      CGT R      AGT S      GGT G
# TGC C      CGC R      AGC S      GGC G
# TGA Stop   CGA R      AGA R      GGA G
# TGG W      CGG R      AGG R      GGG G"""
# table = dict(zip(table.split()[::2],table.split()[1::2]))


def parse_sequence_into_translated_regions(sequence, starts=["ATG", "CTG", "GTG"], min_length=10, reinitiation=False):
    '''
    take a nucleotide sequence and return a list of ORFs
    '''
    frames = {1:[], 2:[], 3:[]}
    for i in range(len(sequence)+1):
        try:
            codon = sequence[i:i+3]
            if len(codon) == 3:
                frames[1].append(codon)
        except KeyError:
            continue
        try:
            codon = sequence[i+1:i+4]
            if len(codon) == 3:
                frames[1].append(codon)
        except KeyError:
            continue
        try:
            codon = sequence[i+2:i+5]
            if len(codon) == 3:
                frames[1].append(codon)
        except KeyError:
            continue
    
    stops = ["TAA","TAG","TGA"]
    

    orf_sequences = [] 
    for i in range(len(sequence)):
        if sequence[i:i+3] in starts:
            for j in range(i,len(sequence),3):
                if sequence[j:j+3] in stops:
                    orf = [sequence[k:k+3] for k in range(i,j+3,3)]
                    if len(''.join(orf)) > min_length:
                        orf_sequences.append(''.join(orf))
                    break

    orfs = [] 
    counter = 1
    for orf in orf_sequences:
        start_codon_position = sequence.find(orf)
        stop_codon_position = sequence.find(orf) + len(orf)
        orfs.append((start_codon_position, stop_codon_position))
        counter += 1
    return orfs


def build_graphs_from_fasta(file_path, min_lenth=100, start_codons=["ATG", "CTG", "GTG"], reinitiation=False, readthrough=False, upstream_limit=1):
    file = open(file_path, 'r').readlines()
    sequences = {} 
    for line in file:
        if line[0] == '>':
            name = line.split(" ")[0][1:]
            sequences[name] = ""
            
        else:
            sequences[name] += line.strip('\n')

    graphs = []
    for sequence in sequences:
        orfs = parse_sequence_into_translated_regions(sequences[sequence], starts=start_codons, min_length=min_lenth)
        dg = RDG(name=sequence, locus_stop=len(sequences[sequence]))
        with Bar('building...') as bar:
            for orf in sorted(orfs):
                dg.add_open_reading_frame(orf[0], orf[1], reinitiation=reinitiation)

                bar.next()
        graphs.append(dg)
    return graphs

if __name__ == "__main__":
    # graphs = build_graphs_from_fasta('/home/jack/projects/decision_graphs/data/PHPT1_transcript_sequence.fa', min_lenth=0)
    graphs = build_graphs_from_fasta('/home/jack/projects/decision_graphs/data/test_fasta_multi_sequences.fa')
    for dg in graphs:
        plot(dg)




from RDG import RDG, Node, Edge, save, load
# from RDG_to_file import save, load

def test_read_write():
    dg = RDG()
    dg.add_open_reading_frame(30, 90)
    dg.add_open_reading_frame(131, 171)
    dg.add_open_reading_frame(150,850)
   
    save(dg, 'test_output.sqlite')
    dg2 = load('', 'test_output.sqlite')

    assert dg.describe() == dg2.describe()



from RDG import RDG, Node, Edge


def test_load_locus_stop_update():
    g = RDG()
    g = RDG.Load(g, locus_name="GeneA", locus_start=0, locus_stop=10000)
    assert g.locus_stop == 10000


def test_load_nodes():
    g = RDG()
    g = RDG.load_example(g)
    assert len(list(g.nodes.keys())) == 5


import matplotlib.pyplot as plt
import networkx as nx
from RDG import RDG, Node, Edge
from matplotlib.gridspec import GridSpec



def position_graphs_nodes(graph):
    '''
    determine the positioning of rdg nodes from graph structure. 
    This is a hacky method that determines Y axis by node key rather than calculating an optimum based on overlaps
    '''

    node_x_positions = [(graph.nodes[node].key, graph.nodes[node].node_start) for node in graph.nodes]
    pos = {}
    for i in node_x_positions:
        pos[i[0]] = (i[1], i[0])
    
    highest_node = max(list(pos.keys()))
    pos[2] = (pos[2][0], pos[highest_node][1] + 1)

    return pos


default_color_dict = {
    'edge_colors':{
        'frame0':(1,0,0),
        'frame1':(0,1,0),
        'frame2':(0,0,1)
    },
    'node_colors':{
        'startpoint':(0,0,1),
        'endpoint':(0.5,0,0.5),
        'translation_start':(0,1,0),
        'translation_stop':(1,0,0),
        'frameshift':(1,0.5,0.3)
    }
}



def plot(graph, color_dict=default_color_dict, node_size=10, height_ratios=[2, 1]):
    G = nx.DiGraph()
    edges = graph.get_edges_from_to()

    pos = position_graphs_nodes(graph)

    endpoints = graph.get_endpoints()
    startpoints = graph.get_startpoints()
    translation_starts = graph.get_start_nodes()
    translation_stops = graph.get_stop_nodes()
    frameshifts = graph.get_frameshifts()

    G.add_edges_from(edges.keys())
    node_colors = []
    for node in G.nodes():
        if node in startpoints:
            node_colors.append(default_color_dict['node_colors']['startpoint'])
        elif node in endpoints:
            node_colors.append(default_color_dict['node_colors']['endpoint'])
        elif node in translation_starts:
            node_colors.append(default_color_dict['node_colors']['translation_start'])
        elif node in translation_stops:
            node_colors.append(default_color_dict['node_colors']['translation_stop'])
        elif node in frameshifts:
            node_colors.append(default_color_dict['node_colors']['frameshift'])
        else:
            node_colors.append((0,0,0))
    
    edge_colors = []
    for edge in G.edges:
        if graph.edges[edges[edge]].edge_type == "translated":
            frame = graph.edges[edges[edge]].frame
        else:
            frame = None

        if frame == 0:
            edge_colors.append(default_color_dict['edge_colors']['frame0'])
        elif frame == 1:
            edge_colors.append(default_color_dict['edge_colors']['frame1'])
        elif frame == 2:
            edge_colors.append(default_color_dict['edge_colors']['frame2'])
        else:
            edge_colors.append((0,0,0))

    fig = plt.figure()
    fig.suptitle(f"Visualisation of an RDG for {graph.locus}")

    gs = GridSpec(2, 1, width_ratios=[1], height_ratios=height_ratios, hspace=0)
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])

    nx.draw_networkx(G, pos=pos, ax=ax1, node_shape='o', node_size=node_size, node_color=node_colors, edge_color=edge_colors, with_labels=False)

    orfs_in_frame = {0:[], 1:[], 2:[]} 

    orfs = graph.get_orfs()
    for orf in orfs: 
        if len(orf) == 2:
            orfs_in_frame[orf[0]%3].append((orf[0], orf[1] - orf[0]))
        elif len(orf) == 3:
            orfs_in_frame[orf[0]%3].append((orf[0], orf[1] - orf[0]))
            orfs_in_frame[orf[1]%3].append((orf[1], orf[2] - orf[1]))



    height = 10
    yticks_heights = []
    yticks_labels = []
    for frame, color in zip(sorted(orfs_in_frame.keys()), [(1,0,0), (0,1,0),  (0,0,1)]):
        yticks_heights.append(height + 1)
        yticks_labels.append("Frame " + str(frame + 1))
        ax2.broken_barh(sorted(orfs_in_frame[frame]), (height, 2), facecolors=color, edgecolor='black',)
        height += 3

    ax2.tick_params(bottom=True, labelbottom=True)
    ax2.set_xlim(left=ax1.get_xlim()[0], right=ax1.get_xlim()[1])
    ax2.set_ylim(bottom=10, top=19)
    # ax2.set_yticks(yticks_heights, labels=yticks_labels)

    # plt.show()
    return fig, ax1, ax2


# if __name__ == "__main__":
#     dg = RDG(name="Example Gene")
#     dg.add_open_reading_frame(30, 90)
#     dg.add_open_reading_frame(61, 400)
#     dg.add_open_reading_frame(92, 150)

#     # dg.add_open_reading_frame(550, 850)
#     # dg.add_stop_codon_readthrough(850, 880)
#     dg.add_frameshift(400, 450, 2)
#     # print(dg.get_orfs())
#     plot(dg)



from RDG import RDG, Node, Edge


def test_update_edge_input_to_new_to_node():
    g = RDG()
    g = RDG.load_example(g)
    node_key = g.get_new_node_key()

    start_codon_position = 15
    stop_codon_position = 25
    start_node = Node(key=node_key, node_type="start_codon", coordinates=start_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])

    node_key = node_key + 1
    stop_node = Node(key=node_key, node_type="stop_codon", coordinates=stop_codon_position, edges_in=[], edges_out=[], nodes_in=[], nodes_out=[])
    
    g.update_edge(2, 1, 5, (1,1000))
    tf = 2 in g.edges
    assert 2 in g.nodes[5].input_edges

def test_update_edge_input_to_old_to_node():
    g = RDG()
    g = RDG.load_example(g)
    g.update_edge(2, 1, 5, (1,1000))
    assert 2 not in g.nodes[2].input_edges


def test_update_edge_input_to_new_from_node():
    g = RDG()
    g = RDG.load_example(g)

    g.update_edge(2, 1, 5, (1,1000))
    assert 2 in g.nodes[1].output_edges


def test_update_edge_input_to_old_from_node():
    g = RDG()
    g = RDG.load_example(g)
    g.update_edge(2, 1, 5, (1,1000))
    assert 2 not in g.nodes[3].output_edges


def test_update_edge_input_to_new_to_node():
    g = RDG()
    g = RDG.load_example(g)

    g.update_edge(2, 1, 5, (1,1000))
    assert 2 in g.nodes[1].output_edges


from RDG import RDG, Node, Edge


def test_basic_edges():
    assert len(RDG().edges) == 1

def test_basic_nodes():
    assert list(RDG().nodes.keys()) == [1, 2]

def test_root_to_node_acyclic_node_path():
    g = RDG()
    g = RDG.load_example(g)
    assert g.root_to_node_of_acyclic_node_path(5) == [1, 3, 4, 5]


from RDG import RDG, Node, Edge

def test_get_vertices():
    g = RDG()
    g = RDG.load_example(g)
    assert g.vertices() == [1,2,3,4,5]

def test_get_edges():
    g = RDG()
    g = RDG.load_example(g)
    assert g.get_edges() == [1,2,3,4]

def test_get_edges_from_to():
    g = RDG()
    g = RDG.load_example(g)
    edges = g.get_edges_from_to()    
    assert sorted(edges) == sorted([(1,3),(3,4), (4,5), (3,2)])

def test_get_startpoints():
    g = RDG()
    g = RDG.load_example(g)
    startpoints = g.get_startpoints()
    assert len(startpoints) == 1



from RDG import RDG, Node, Edge


def test_remove_edge_from_node():
    g = RDG()
    g = RDG.load_example(g)
    g.remove_edge(2)
    assert g.nodes[3].output_edges == [3]

def test_remove_edge_to_node():
    g = RDG()
    g = RDG.load_example(g)
    g.remove_edge(2)
    assert g.nodes[2].input_edges == []

def test_remove_edge_edges():
    g = RDG()
    g = RDG.load_example(g)
    g.remove_edge(2)
    tf = 2 in g.edges
    assert tf == False